@@ -1,6 +1,4 @@
 from collections import Counter
-import cPickle as pickle
-import datetime
 import json
 import logging
 import os
@@ -8,18 +6,16 @@ import os.path
 import pprint
 import signal
 from subprocess import call
-import sys
 from time import sleep
 import traceback
 
-import arrow
 import boto.swf.layer2 as swf
 import daemon
 import luigi.configuration
 from six import iteritems
 
 from .util import default_log_format, dthandler, kill_from_pid_file, \
-    SingleWaitingLockPidFile, get_all_tasks, get_class
+    SingleWaitingLockPidFile, get_task_configurations, get_class, dt_from_iso
 
 
 logger = logging.getLogger(__name__)
@@ -37,6 +33,114 @@ attrWfSignaled = 'workflowExecutionSignaledEventAttributes'
 seconds = 1.
 
 
+class WfState(object):
+
+    def read_wf_state(self, events, task_configs):
+        # Read what has happened.
+        self.version = self._get_version(events)
+        self.schedulings = self._get_all_schedulings(events)
+        self.completed = self._get_completed(events, task_configs)
+        self.failures = self._get_failures(events)
+        self.timeouts = self._get_timeouts(events)
+        self.retries = self._get_retries(events)
+        self.wf_cancel_req = self._get_wf_cancel_requested(events)
+        self.cancel_requests = self._get_task_cancel_requests(events)
+        self.cancellations = self._get_activity_cancellations(events)
+        # Make inferences.
+        self.running = self._get_running()
+
+    def _get_version(self, events):
+        wf_event = next(e for e in events
+                        if e['eventType'] == 'WorkflowExecutionStarted')
+        wf_event_attr = wf_event['workflowExecutionStartedEventAttributes']
+        return wf_event_attr['workflowType']['version']
+
+    def _get_task_id(self, events, event_attributes):
+        event_id = event_attributes['scheduledEventId'] - 1
+        activity = events[event_id]
+        attributes = activity['activityTaskScheduledEventAttributes']
+        return attributes['activityId']
+
+    def _get_all_schedulings(self, events):
+        all_schedulings = \
+            [e[attrTaskScheduled]['activityId']
+             for e in events
+             if e['eventType'] == 'ActivityTaskScheduled']
+        return Counter(all_schedulings)
+
+    def _get_completed(self, events, task_configs):
+        # Get completed activity tasks.
+        compl_act = [self._get_task_id(events, e[attrTaskCompleted])
+                     for e in events
+                     if e['eventType'] == 'ActivityTaskCompleted']
+        # Get completed wrappers.
+        # Iterate so wrappers can depend on one another.
+        completed_wrappers = []
+        while True:
+            completed = compl_act + completed_wrappers
+            new_comp = [task_id
+                        for task_id, task in iteritems(task_configs)
+                        if (task['is_wrapper']
+                            and task_id not in completed
+                            and all(t in completed for t in task['deps']))]
+            if len(new_comp) == 0:
+                break
+            completed_wrappers += new_comp
+        return compl_act + completed_wrappers
+
+    def _get_failures(self, events):
+        return Counter([self._get_task_id(events, e[attrTaskFailed])
+                        for e in events
+                        if e['eventType'] == 'ActivityTaskFailed'])
+
+    def _get_timeouts(self, events):
+        return Counter([self._get_task_id(events, e[attrTaskTimedOut])
+                        for e in events
+                        if e['eventType'] == 'ActivityTaskTimedOut'])
+
+    def _get_retries(self, events):
+        def parse_retries(task_ids):
+            try:
+                return map(lambda s: s.strip(), task_ids.strip().split('\n'))
+            except:
+                tb = traceback.format_exc()
+                logger.error('get_wf_state_full._get_retries'
+                             '.parse_retries():\n%s', tb)
+                return []
+
+        retries_flat = \
+            [retry
+             for e in events
+             if (e['eventType'] == 'WorkflowExecutionSignaled'
+                 and e[attrWfSignaled]['signalName'] == 'retry')
+             for retry in parse_retries(e[attrWfSignaled]['input'])]
+        return Counter(retries_flat)
+
+    def _get_wf_cancel_requested(self, events):
+        return any(e for e in events
+                   if e['eventType'] == 'WorkflowExecutionCancelRequested')
+
+    def _get_task_cancel_requests(self, events):
+        return [e[attrTaskCancReq]['activityId']
+                for e in events
+                if e['eventType'] == 'ActivityTaskCancelRequested']
+
+    def _get_activity_cancellations(self, events):
+        return Counter([self._get_task_id(events, e[attrTaskCanceled])
+                        for e in events
+                        if e['eventType'] == 'ActivityTaskCanceled'])
+
+    def _get_running(self):
+        return \
+            [task_id
+             for task_id, sched_count
+             in iteritems(self.schedulings)
+             if (task_id not in self.completed and
+                 sched_count > (self.failures[task_id]
+                                + self.timeouts[task_id]
+                                + self.cancellations[task_id]))]
+
+
 class LuigiSwfDecider(swf.Decider):
     """Implementation of boto's SWF Decider
 
@@ -60,14 +164,13 @@ class LuigiSwfDecider(swf.Decider):
                 logger.debug('LuigiSwfDecider().run(), poll timed out')
                 return
             events = self._get_events(decision_task)
-            state = self._get_state(events)
-            if not state['wf_cancel_req']:
-                version = self._get_version(events)
-                self._schedule_activities(state, decisions, version)
-            else:
-                self._cancel_activities(state, decisions)
+            task_configs = self._get_task_configurations(
+                events, decision_task['workflowExecution']['runId'])
+            state = WfState()
+            state.read_wf_state(events, task_configs)
+            self._decide(state, decisions, task_configs)
             self.complete(decisions=decisions)
-        except Exception, error:
+        except Exception as error:
             tb = traceback.format_exc()
             reason = ('Decider failed:\n' + str(error))[:255]
             details = (tb + '\n' + str(error))[:32767]
@@ -77,6 +180,12 @@ class LuigiSwfDecider(swf.Decider):
             self.complete(decisions=decisions)
             raise
 
+    def _decide(self, state, decisions, task_configs):
+        if not state.wf_cancel_req:
+            self._schedule_activities(state, decisions, task_configs)
+        else:
+            self._cancel_activities(state, decisions)
+
     def _get_events(self, decision_task):
         # It's paginated.
         events = decision_task['events']
@@ -87,172 +196,41 @@ class LuigiSwfDecider(swf.Decider):
                 events += decision_task['events']
         return events
 
-    def _get_all_tasks(self, events):
-        wf_event = next(e for e in events
-                        if e['eventType'] == 'WorkflowExecutionStarted')
-        wf_event_attr = wf_event['workflowExecutionStartedEventAttributes']
-        wf_input = json.loads(str(wf_event_attr['input']))
-        wf_params = wf_input['wf_params']
-        logger.debug('LuigiSwfDecider()._get_all_tasks(), wf_input:\n%s',
-                     pp.pformat(wf_input))
-        wf_cls = get_class(*wf_input['wf_task'])
-        kwargs = dict()
-        for param_name, param_cls in wf_cls.get_params():
-            if param_name == 'pool':
-                continue
-            if isinstance(param_cls, luigi.DateParameter):
-                kwargs[param_name] = arrow.get(wf_params[param_name])
-            else:
-                kwargs[param_name] = wf_params[param_name]
-        wf_task = wf_cls(**kwargs)
-        all_tasks = get_all_tasks(wf_task, include_obj=True)
-        return all_tasks
-
-    def _get_version(self, events):
-        wf_event = next(e for e in events
-                        if e['eventType'] == 'WorkflowExecutionStarted')
-        wf_event_attr = wf_event['workflowExecutionStartedEventAttributes']
-        return wf_event_attr['workflowType']['version']
-
-    def _get_task_id(self, events, event_attributes):
-        event_id = event_attributes['scheduledEventId'] - 1
-        activity = events[event_id]
-        attributes = activity['activityTaskScheduledEventAttributes']
-        return attributes['activityId']
-
-    def _get_runnables(self, all_tasks, completed, running):
-        """
-        >>> all_tasks = {
-        ...     'Task1': {'deps': ['Task3']},
-        ...     'Task2': {'deps': ['Task3']},
-        ...     'Task3': {'deps': []},
-        ...     'Task4': {'deps': []},
-        ... }
-        >>> completed = ['Task3']
-        >>> running = ['Task4']
-        >>> decider = LuigiSwfDecider()
-        >>> list(decider._get_runnables(all_tasks, completed, running))
-        [('Task1', {'deps': ['Task3']}), ('Task2', {'deps': ['Task3']})]
-        >>> completed = ['Task1']
-        >>> list(decider._get_runnables(all_tasks, completed))
-        [('Task3', {'deps': []})]
-        """
-        for task_id, task in iteritems(all_tasks):
-            if task_id not in completed and task_id not in running and \
-                    all(d in completed for d in task['deps']):
-                yield task_id, task
-
-    def _get_unretryables(self, all_tasks, failed, retries):
-        for task_id, count in iteritems(failed):
-            if count > all_tasks[task_id]['retries'] + retries.get(task_id, 0):
-                yield task_id
-
-    def _get_completed_wrappers(self, all_tasks, completed_activities):
-        # Iterate so wrappers can depend on one another.
-        completed_wrappers = []
-        while True:
-            completed = completed_activities + completed_wrappers
-            new_comp = [task_id
-                        for task_id, task in iteritems(all_tasks)
-                        if (task['is_wrapper']
-                            and task_id not in completed
-                            and all(t in completed for t in task['deps']))]
-            if len(new_comp) == 0:
-                break
-            completed_wrappers += new_comp
-        return completed_wrappers
-
-    def _get_state(self, events):
-        all_tasks = self._get_all_tasks(events)
-        state = dict()
-        state['all_schedulings'] = \
-            [e[attrTaskScheduled]['activityId']
-             for e in events
-             if e['eventType'] == 'ActivityTaskScheduled']
-        state['schedulings'] = Counter(state['all_schedulings'])
-        state['compl_act'] = [self._get_task_id(events, e[attrTaskCompleted])
-                              for e in events
-                              if e['eventType'] == 'ActivityTaskCompleted']
-        state['compl_w'] = self._get_completed_wrappers(all_tasks,
-                                                        state['compl_act'])
-        state['completed'] = state['compl_act'] + state['compl_w']
-        failures_flat = [self._get_task_id(events, e[attrTaskFailed])
-                         for e in events
-                         if e['eventType'] == 'ActivityTaskFailed']
-        state['failures'] = Counter(failures_flat)
-        timeouts_flat = [self._get_task_id(events, e[attrTaskTimedOut])
-                         for e in events
-                         if e['eventType'] == 'ActivityTaskTimedOut']
-        state['timeouts'] = Counter(timeouts_flat)
-        retries_flat = \
-            [retry
-             for e in events
-             if (e['eventType'] == 'WorkflowExecutionSignaled'
-                 and e[attrWfSignaled]['signalName'] == 'retry')
-             for retry in self._parse_retries(e[attrWfSignaled]['input'])]
-        state['retries'] = Counter(retries_flat)
-        state['wf_cancel_req'] = any(e for e in events
-                                     if e['eventType'] == ('WorkflowExecution'
-                                                           'CancelRequested'))
-        state['cancel_requests'] = [e[attrTaskCancReq]['activityId']
-                                    for e in events
-                                    if e['eventType'] == ('ActivityTask'
-                                                          'CancelRequested')]
-        act_cancels = [self._get_task_id(events,
-                                         e[attrTaskCanceled])
-                       for e in events
-                       if e['eventType'] == ('ActivityTask'
-                                             'Canceled')]
-        state['cancellations'] = Counter(act_cancels)
-        state['running'] = \
-            [task_id
-             for task_id, sched_count
-             in iteritems(state['schedulings'])
-             if (task_id not in state['completed'] and
-                 sched_count > (state['failures'][task_id]
-                                + state['timeouts'][task_id]
-                                + state['cancellations'][task_id]))]
-        state['runnables'] = list(self._get_runnables(all_tasks,
-                                                      state['completed'],
-                                                      state['running']))
-        state['unretryables'] = list(self._get_unretryables(all_tasks,
-                                                            state['failures'],
-                                                            state['retries']))
-        logger.debug('LuigiSwfDecider().get_state(), state:\n%s',
-                     pp.pformat(state))
-        return state
-
-    def _parse_retries(self, task_ids):
-        try:
-            return map(lambda s: s.strip(), task_ids.strip().split('\n'))
-        except:
-            tb = traceback.format_exc()
-            logger.error('LuigiSwfDecider()._parse_retries():\n%s', tb)
-            return []
-
-    def _schedule_activities(self, state, decisions, version):
+    def _schedule_activities(self, state, decisions, task_configs):
         scheduled_count = 0
-        for task_id, task in state['runnables']:
-            if task_id in state['unretryables']:
+        runnables = self._get_runnables(state, task_configs)
+        running_mutexes = self._get_running_mutexes(state, task_configs)
+        unretryables = self._get_unretryables(state, task_configs)
+        for task_id, task in iteritems(runnables):
+            if task_id in unretryables:
                 continue
+            if task['running_mutex'] is not None:
+                # These tasks want to run one-at-a-time per workflow execution.
+                if task['running_mutex'] in running_mutexes:
+                    continue
+                running_mutexes.add(task['running_mutex'])
             scheduled_count += 1
             start_to_close = task['start_to_close_timeout']
             schedule_to_start = task['schedule_to_start_timeout']
             schedule_to_close = task['schedule_to_close_timeout']
+            inp = {
+                'class': task['class'],
+                'params': task['params'],
+            }
             decisions.schedule_activity_task(
                 activity_id=task_id,
                 activity_type_name=task['task_family'],
-                activity_type_version=version,
+                activity_type_version=state.version,
                 task_list=task['task_list'],
                 heartbeat_timeout=str(task['heartbeat_timeout']),
                 start_to_close_timeout=str(start_to_close),
                 schedule_to_start_timeout=str(schedule_to_start),
                 schedule_to_close_timeout=str(schedule_to_close),
-                input=json.dumps(task, default=dthandler))
+                input=json.dumps(inp, default=dthandler))
             logger.debug('LuigiSwfDecider().run(), scheduled %s', task_id)
-        if scheduled_count == 0 and len(state['running']) == 0:
-            if len(state['unretryables']) > 0:
-                msg = 'Task(s) failed: ' + ', '.join(state['unretryables'])
+        if scheduled_count == 0 and len(state.running) == 0:
+            if len(unretryables) > 0:
+                msg = 'Task(s) failed: ' + ', '.join(unretryables)
                 logger.error('LuigiSwfDecider().run(), '
                              'failing execution: %s', msg)
                 decisions.fail_workflow_execution(reason=msg[:255],
@@ -263,13 +241,56 @@ class LuigiSwfDecider(swf.Decider):
                 decisions.complete_workflow_execution()
 
     def _cancel_activities(self, state, decisions):
-        if len(state['running']) > 0:
-            for task_id in state['running']:
-                if task_id not in state['cancel_requests']:
+        if len(state.running) > 0:
+            for task_id in state.running:
+                if task_id not in state.cancel_requests:
                     decisions.request_cancel_activity_task(task_id)
         else:
             decisions.cancel_workflow_executions()
 
+    def _get_runnables(self, state, task_configs):
+        result = {}
+        for task_id, task in iteritems(task_configs):
+            if task_id not in state.completed and \
+                    task_id not in state.running and \
+                    all(d in state.completed for d in task['deps']):
+                result[task_id] = task
+        return result
+
+    def _get_unretryables(self, state, task_configs):
+        result = []
+        for task_id, count in iteritems(state.failures):
+            if count > task_configs[task_id]['retries'] + \
+                    state.retries.get(task_id, 0):
+                result.append(task_id)
+        return result
+
+    def _get_task_configurations(self, events, wf_run_id):
+        wf_event = next(e for e in events
+                        if e['eventType'] == 'WorkflowExecutionStarted')
+        wf_event_attr = wf_event['workflowExecutionStartedEventAttributes']
+        wf_input = json.loads(str(wf_event_attr['input']))
+        wf_params = wf_input['wf_params']
+        logger.debug('get_wf_state_full.get_task_configurations(), '
+                     'wf_input:\n%s', pp.pformat(wf_input))
+        wf_cls = get_class(*wf_input['wf_task'])
+        kwargs = dict()
+        for param_name, param_cls in wf_cls.get_params():
+            if param_name == 'pool':
+                continue
+            if isinstance(param_cls, luigi.DateParameter):
+                kwargs[param_name] = dt_from_iso(wf_params[param_name])
+            else:
+                kwargs[param_name] = wf_params[param_name]
+        wf_task = wf_cls(**kwargs)
+        return get_task_configurations(wf_task, wf_run_id, include_obj=True)
+
+    def _get_running_mutexes(self, state, task_configs):
+        return \
+            set(task['running_mutex']
+                for task_id, task in iteritems(task_configs)
+                if task_id in state.running and task['running_mutex'])
+
 
 class DeciderServer(object):
     """Decider daemon
@@ -389,7 +410,7 @@ class DeciderServer(object):
                 try:
                     logger.debug('DeciderServer().start(), decider.run()')
                     decider.run(self.identity)
-                except Exception as ex:
+                except Exception:
                     tb = traceback.format_exc()
                     logger.error('DeciderServer().start(), error:\n%s', tb)
                 sleep(0.001 * seconds)
